/* 
Anpassung des Programmes "Pico-Fmeter2a" auf Arduino-IDE.

2024-02-26
Michael Nowak
http://www.mino-elektronik.de
Alle Angaben ohne Gewaehr !
*/

#define PICO_FMETER2
#include "System-Pico-Fmeter2.h"

#define VERSION_LCD1 "mino PicoFMeter2"
#define VERSION_LCD2 "V3.a  2024-02-26"
#define VERSION_SER  "mino Pico-FMeter V3.0 2024-02-26"

uint32_t *irq_tab;                          // fuer IRQ-Vektoren im RAM

char version[] = VERSION_SER;               // fuer serielle Ausgabe

uint32_t SystemCoreClock  = (DEF_SYSCLOCK); // 133 MHz
uint32_t F_xtal = MHZ_12;                   // onboard default

double SM0_takt;                            // Referenz mit Poti korrigiert
uint32_t ms_counter;                        // fuer freie Verwendung
volatile uint8_t ms_flag, adc_flag;         // ms-flags
uint8_t ex_ref_aktiv,
        sprache;


void kurz_warten(void)                      // fuer kurze Verzoegerungen ohne IRQ
{
  volatile int i = 1000;
  while(i--);
}  

// POSTDIV2 / 2 => SystemCoreClock *= 2
void set_turbo_mode()
{
uint32_t div1, div2, ref_div;
  VREG_AND_CHIP_RESET_SET-> VREG = 15<<VREG_AND_CHIP_RESET_VREG_VSEL_Pos;
  while(!(VREG_AND_CHIP_RESET->VREG & VREG_AND_CHIP_RESET_VREG_ROK_Msk));
  kurz_warten();       // Vreg abwarten
  div1 = (PLL_SYS->PRIM >> PLL_SYS_PRIM_POSTDIV1_Pos) & 0x07;
  div2 = (PLL_SYS->PRIM >> PLL_SYS_PRIM_POSTDIV2_Pos) & 0x07;
  div2 /= 2;
  ref_div = (PLL_SYS->CS >> PLL_SYS_CS_REFDIV_Pos) & 0x3f;    // PLL-Vorteiler

  PLL_SYS->PRIM = (div1 << PLL_SYS_PRIM_POSTDIV1_Pos) | 
                  (div2 << PLL_SYS_PRIM_POSTDIV2_Pos);
  while (!(PLL_SYS->CS & PLL_SYS_CS_LOCK_Msk));
                  
  SystemCoreClock = PLL_SYS->FBDIV_INT * (F_xtal/ref_div) / (div1 * div2);
} 

// LCD_RS und LCD_RW auf 3k3 Widerstand testen und ggf. SystemCoreClock verdoppeln
int teste_turbo(void)
{
uint8_t turbo = 1;                          // Vorgabewert  
  RESETS_CLR->RESET = RESETS_RESET_io_bank0_Msk | 
                      RESETS_RESET_pads_bank0_Msk;
  while((RESETS->RESET_DONE & (RESETS_RESET_io_bank0_Msk | RESETS_RESET_pads_bank0_Msk)) !=
        (RESETS_RESET_io_bank0_Msk | RESETS_RESET_pads_bank0_Msk));
  IO_BANK0->GPIO1_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_sio_0;
  IO_BANK0->GPIO2_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_sio_0;
  PADS_BANK0->GPIO1 = PADS_BANK0_GPIO0_IE_Msk | PADS_BANK0_GPIO0_PUE_Msk;  // nur pullup = 1
  PADS_BANK0->GPIO2 = PADS_BANK0_GPIO0_IE_Msk | PADS_BANK0_GPIO0_PUE_Msk;  // nur pullup = 1
  SIO->GPIO_OE_SET =  BIT(2);               // GPIO2 = 0
  kurz_warten();
  if((SIO->GPIO_IN & BIT(1))) turbo = 0;    // GPIO1 bleibt oben: LCD_RW ist nicht verbunden
  SIO->GPIO_OE_CLR =  BIT(2);               // GPIO2 = Eingang
  SIO->GPIO_OE_SET =  BIT(1);               // GPIO1 = 0
  kurz_warten();
  if((SIO->GPIO_IN & BIT(2))) turbo = 0;    // GPIO2 bleibt oben: LCD_RW ist nicht verbunden
  SIO->GPIO_OUT_CLR =  BIT(1);              // GPIO1 = Eingang
  return(turbo);                            // turbo-modus angefordert
}  


// alle Millisekunden aufrufen
void TIMER_IRQ_0_IRQHandler(void)
{
  TIMER->INTR = TIMER_INTR_ALARM_0_Msk;
  TIMER->ALARM0 = TIMER->TIMELR + MS_TEILER;
  ms_counter++;
  ms_flag = 1;
  adc_flag = 1;
}

// Timer ist schon von USB (Kanal 3) initialisiert
void init_timer(void)
{
  RESETS_CLR->RESET = RESETS_RESET_timer_Msk;     // ist schon aktiv, stoert aber nicht
  while (!(RESETS->RESET_DONE & RESETS_RESET_timer_Msk));
  irq_tab[TIMER_IRQ_0_IRQn] = (uint32_t)TIMER_IRQ_0_IRQHandler; // Vektor ins RAM

  NVIC_EnableIRQ(TIMER_IRQ_0_IRQn);
  NVIC_SetPriority(TIMER_IRQ_0_IRQn,3);           // geringe Prioritaet
  TIMER->ALARM0 = TIMER->TIMELR + MS_TEILER*100;
  TIMER->INTE |= TIMER_INTE_ALARM_0_Msk;
}

void warten_ms(uint32_t ms)
{
  do {
    while(!ms_flag);
    ms_flag = 0;
  } while(--ms);
} 


void init_adc0(void)                              // ADC-Kanal0 dauernd wandeln
{
  RESETS_CLR->RESET = RESETS_RESET_adc_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_adc_Msk));
  ADC->CS = ADC_CS_START_MANY_Msk | ADC_CS_EN_Msk;
  PADS_BANK0->GPIO26 = PADS_BANK0_GPIO0_OD_Msk;   // output disable
}


uint32_t lese_poti(void)    // aktuellen Wert lesen
{
  return(ADC->RESULT);      // ADC0 an GPIO26
}  

// auf externen Trimmer testen und ggf. zur Frequenzkorrektur verwenden
void teste_abgleich(void)
{
static int32_t summe = ADC_MAX/2 * ADC_FILTER;        // Startwert bei +/- 0
static uint32_t poti_vorhanden, poti_init;
double f_ref_korrektur = 0.0;                         // kein Poti bestueckt
  if(!poti_init) {
    poti_init = 1;
    IO_BANK0->GPIO26_CTRL = IO_BANK0_GPIO0_CTRL_FUNCSEL_sio_0;  // GPIO
    PADS_BANK0->GPIO26 = PADS_BANK0_GPIO0_IE_Msk | PADS_BANK0_GPIO0_PUE_Msk;  // nur pullup = 1
    kurz_warten();
    if(lese_poti() <= 4000) poti_vorhanden = 2;       // pullup zieht nicht ganz hoch
    PADS_BANK0->GPIO26 = PADS_BANK0_GPIO0_IE_Msk | PADS_BANK0_GPIO0_PDE_Msk;  // nur pulldown = 1
    kurz_warten();
    if(lese_poti() >= 100) poti_vorhanden = 3;        // pulldown zieht nicht ganz runter
    PADS_BANK0->GPIO26 = PADS_BANK0_GPIO0_IE_Msk;     // pullup = pulldown = 0 als ADC-Eingang
  }  

  if(adc_flag && poti_vorhanden) {
    adc_flag = 0;
    summe -= summe/ADC_FILTER;
    summe += (lese_poti() - ADC_MAX/2);               // +/- 0 ppm in Mittelstellung
// Korrekturwert von +/- 2047 auf +/- 20,47 ppm von SystemCoreClock abbilden
    f_ref_korrektur = (double)SystemCoreClock * (double)summe/ADC_FILTER*1e-8;;
  }
  SM0_takt = (SystemCoreClock+f_ref_korrektur)/PIO_TIME_DIV;  
}

// XOSC aktivieren, SystemCoreClock setzen und Taktsignale aktivieren
void init_clock(void)
{
volatile int i;

  XOSC->CTRL     = (XOSC_CTRL_FREQ_RANGE_1_15MHZ << XOSC_CTRL_FREQ_RANGE_Pos);
  XOSC->STARTUP  = 100;             // ca. 2 ms bei 12 MHz
  XOSC_SET->CTRL = (XOSC_CTRL_ENABLE_ENABLE << XOSC_CTRL_ENABLE_Pos);
  while (!(XOSC->STATUS & XOSC_STATUS_STABLE_Msk));
  
  RESETS_CLR->RESET = RESETS_RESET_pll_sys_Msk;
  while (!(RESETS->RESET_DONE & RESETS_RESET_pll_sys_Msk));


  PLL_SYS->CS = (REF_DIV << PLL_SYS_CS_REFDIV_Pos); // PLL-Aufloesung vorgeben
  PLL_SYS->FBDIV_INT = (SystemCoreClock*PLL_DIV2*REF_DIV)/(F_xtal/PLL_DIV1);  // PLL-VCO typ. 1310 MHz
// eff. Taktfrequenz ermitteln  
  PLL_SYS->PRIM = (PLL_DIV1 << PLL_SYS_PRIM_POSTDIV1_Pos) | 
                  (PLL_DIV2 << PLL_SYS_PRIM_POSTDIV2_Pos);
  SystemCoreClock = PLL_SYS->FBDIV_INT *(F_xtal/REF_DIV) / 
                    (PLL_DIV1*PLL_DIV2);

  PLL_SYS_CLR->PWR = PLL_SYS_PWR_VCOPD_Msk | PLL_SYS_PWR_PD_Msk;
  while (!(PLL_SYS->CS & PLL_SYS_CS_LOCK_Msk));

  PLL_SYS_CLR->PWR = PLL_SYS_PWR_POSTDIVPD_Msk;

  CLOCKS->CLK_REF_CTRL = (CLOCKS_CLK_REF_CTRL_SRC_xosc_clksrc << CLOCKS_CLK_REF_CTRL_SRC_Pos);
  CLOCKS->CLK_SYS_CTRL = (CLOCKS_CLK_SYS_CTRL_AUXSRC_clksrc_pll_sys << CLOCKS_CLK_SYS_CTRL_AUXSRC_Pos) |
                         (CLOCKS_CLK_SYS_CTRL_SRC_clksrc_clk_sys_aux << CLOCKS_CLK_SYS_CTRL_SRC_Pos);

  CLOCKS->CLK_PERI_CTRL = CLOCKS_CLK_PERI_CTRL_ENABLE_Msk |
      (CLOCKS_CLK_PERI_CTRL_AUXSRC_clk_sys << CLOCKS_CLK_PERI_CTRL_AUXSRC_Pos);

  CLOCKS->CLK_ADC_CTRL = CLOCKS_CLK_ADC_CTRL_ENABLE_Msk |
      (CLOCKS_CLK_ADC_CTRL_AUXSRC_xosc_clksrc << CLOCKS_CLK_ADC_CTRL_AUXSRC_Pos);

// enable GPIO
  RESETS_CLR->RESET = RESETS_RESET_io_bank0_Msk | 
                      RESETS_RESET_pads_bank0_Msk;
  while((RESETS->RESET_DONE & (RESETS_RESET_io_bank0_Msk | RESETS_RESET_pads_bank0_Msk)) !=
        (RESETS_RESET_io_bank0_Msk | RESETS_RESET_pads_bank0_Msk));
}


void setup()
{
uint8_t ee_init;
int temp;
// Vektor-Tabelle liegt im RAM ab 0x20000000
  irq_tab = (uint32_t *)(PPB->VTOR) + (TIMER_IRQ_0_IRQn - Reset_IRQn + 1);  // Basisadresse +1 wg. stack_ptr

  RESETS_CLR->RESET = RESETS_RESET_io_bank0_Msk | 
                      RESETS_RESET_pads_bank0_Msk;
  while((RESETS->RESET_DONE & (RESETS_RESET_io_bank0_Msk | RESETS_RESET_pads_bank0_Msk)) !=
        (RESETS_RESET_io_bank0_Msk | RESETS_RESET_pads_bank0_Msk));

  temp = teste_turbo();
 // fuer dynamische Taktfrequenz; Arduino kann nur statisch! USB hat eigene PLL  
  init_clock();      
  if(temp) set_turbo_mode();                            // SytemCoreClock *= 2 bei Bedarf
  SM0_takt = SystemCoreClock/PIO_TIME_DIV;              // Basiswert fuer F-messung

  Serial1.begin(DEF_BAUDRATE);                          // USB, Baudrate = dummy
 
// gespeicherte Einstellungen laden; Default-Werte beim 1. Einschalten
  ee_read_uint8(EE_ADR_PARA_INIT, &ee_init);
  if(ee_init == EE_INIT_MUSTER) {                       // EE-Muster ist gueltig
      lese_ee_F1_parameter();                           // dann alte Einstellungen
      lese_ee_F2_parameter();                           // zuruecklesen
      lese_ee_allgemein_parameter();
  } else {
      ee_write_uint8(EE_ADR_PARA_INIT, EE_INIT_MUSTER); // Kennung schreiben
      schreibe_ee_F1_parameter();                       // und danach Defaultwerte
      schreibe_ee_F2_parameter();                       // fuer beide Kanaele
      schreibe_ee_allgemein_parameter();
  }    

  lcd_init();
  init_timer();
  init_adc0();
  init_uart1();

  F1_messung(1);                  // init
  lcd_zeile1((char *)VERSION_LCD1);
  lcd_zeile2((char *)VERSION_LCD2);  
  kurz_warten();
  while(ausgabe_sperre && TIMER->TIMELR <= 2000000) { 
    F1_messung(0);                            // 2 s zum Einschwingen + LCD-Meldung abwarten
  }  
  Serial.print(temp);                         // Status turbo-mode
  Serial.print("  ");            
  Serial.print(PLL_SYS->PRIM, HEX);           // PLL-Einstellung
  Serial.print("  ");            
  Serial.println(SystemCoreClock);            // eff. Taktfrequenz ausgeben

  ausgabe_sperre = 0;
  update_LCD = 1;                             // anzeige freigeben
}

void loop()
{
  int taste;
    F1_messung(0);                              // run
   Fgps_messung();                             // run
   teste_abgleich();
    taste = lese_taste();                       // F1 + Fgps Messungen werden auch hier aufgerufen
    if(taste == PLUS_TASTE) {                   // angezeigte Werte direkt umschalten
      if(anzeige_funktion < MAX_FUNKTION-1) {
        anzeige_funktion++;
        update_LCD = 1;
        ee_write_uint8(EE_ADR_ANZEIGE_FKT, anzeige_funktion);
      }
    } else if(taste == MINUS_TASTE) {           // angezeigte Werte direkt umschalten
      if(anzeige_funktion > 0) {
        anzeige_funktion--;
        update_LCD = 1;
        ee_write_uint8(EE_ADR_ANZEIGE_FKT, anzeige_funktion);
      }
    } else if(taste == PLUS_MINUS) {           // angezeigte Werte direkt umschalten
      if(anzeige_funktion >= FMEAN_UND_N) {
        stat_n = 0;
        update_LCD = 1;
      }
    } else if(taste == INIT_TASTE) {          // Bedienmenue aufrufen
      ausgabe_sperre = 1;                     // Messwertanzeige sperren
      RP2040_hauptmenue();
      ausgabe_sperre = 0;                     // wieder zulassen
      update_LCD = 1;                         // und Anzeige aktualisieren
    }  
}
